dt   = 0.001  # 1ステップあたりの時間間隔
g    = 9.8    # 重力定数

# Ball クラス
class @Ball
  # radius: 半径, mass: 質量, position: 位置, velocity: 速度
  # fulcrumPosition: 振り子の紐の固定点
  constructor: ({@radius, @mass, @position, @velocity, @fulcrumPosition}) ->
    @constrainVelocity()
    @forces =
      gravity: new THREE.Vector3(0, 0, -@mass*g)
      tension: new THREE.Vector3()
    @force = new THREE.Vector3()
    @initializeVerlet(dt)
    @updateValues()
    @elapsedTime = 0

  constrainVelocity: ->
    @velocity.projectOnPlane(@position.clone().sub(@fulcrumPosition))

  initializeVerlet: (dt) ->
    xs = @position.toArray()
    vs = @velocity.toArray()
    as = @accelerations()
    [prevXs, prevVs] = Verlet.prevs(xs, vs, as, dt)
    @prevPosition = new THREE.Vector3().set(prevXs...)
    @prevVelocity = new THREE.Vector3().set(prevVs...)

  updateForce: ->
    relPos = new THREE.Vector3().subVectors(@position, @fulcrumPosition)
    relPosSq = relPos.lengthSq()
    velSq = @velocity.lengthSq()

    # gravity
    @forces.gravity.set(0, 0, -@mass*g)
    gravity = @forces.gravity

    # tension
    alpha = (gravity.dot(relPos) + @mass*velSq)/relPosSq   # Lagrange multiplier
      # Action: I[x] = ∫L[x] dt
      # Lagrangian: L[x] = L(x,v) = m*v^2/2 - m*g*z
      # EoM:
      #   dL/dx_i - (d/dt)*(dL/dv_i) = 0
      # ↓
      # modified Lagrangian: L'(x,v) = L(x,v) - a*g(x)
      # constraint: g(x) = x^2 - l^2
      # modified EoM:
      #   0 = dL'/dx_i - (d/dt)*(dL'/dv_i)
      #     = dL/dx_i - a*(dg/dx_i) - (d/dt)*(dL/dv_i)
    tension = relPos.multiplyScalar(-alpha)
    @forces.tension.copy tension

    @force.copy(gravity).add(tension)

  accelerations: ->
    @updateForce()
    @force.clone().divideScalar(@mass).toArray()

  timeEvolution: (time) ->
    steps = Math.floor(time/dt)
    #steps = 100
    @elapsedTime += steps*dt
    @evolveStep() for [0...steps]

  evolveStep: ->
    xs  = @position.toArray()
    xs_ = @prevPosition.toArray()
    vs_ = @prevVelocity.toArray()
    as  = @accelerations()
    [newXs, newVs] = Verlet.evolve(xs, xs_, vs_, as, dt)

    [@position, @prevPosition] = [@prevPosition, @position]
    [@velocity, @prevVelocity] = [@prevVelocity, @velocity]
    @position.set(newXs...)
    @velocity.set(newVs...)
    @updateValues()

  updateValues: ->
    @x = @position.x
    @y = @position.y
    @z = @position.z
    @v = @velocity.length()
    @vx = @velocity.x
    @vy = @velocity.y
    @vz = @velocity.z
    {@energy, @kinetic, @potential} = @calcEnergy()

  setXyz: ->
    @position.set(@x, @y, @z)
    @constrainVelocity()
    @initializeVerlet(dt)

  setV: ->
    @velocity.set(@vx, @vy, @vz)
    @constrainVelocity()
    @initializeVerlet(dt)

  calcEnergy: ->
    velSq = @velocity.lengthSq()
    k = 1/2*this.mass*velSq
    p = @mass*g*@z
    return {kinetic: k, potential: p, energy: k+p}
